#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>

#include "MainApp.h"
#include "SSocket.h"
#include "ServerRequest.h"

// if you change anything in the code (other than Makefile),
// please change the following string after the hyphen.
#define CLIENTVERSION "2.0e-eolson"

MainApp::MainApp()
{
  // these flags take precedence over any --ctext flag.
  _getopt.addOptionBool(0,"tivo32",false);
  _getopt.addOptionBool(0,"tivo30",false);

  // general options
  _getopt.addOptionBool('h',"help",false);
  _getopt.addOptionInt('v',"verbosity",Logger::MED);
  _getopt.addOptionBool(0,"debug", false);
  _getopt.addOptionString(0,"log-file","");
  _getopt.addOptionBool(0,"silent",false);

  _getopt.addOptionBool('d',"display-progress",true);
  _getopt.addOptionInt(0,"progress-frequency",4);

  // parameters for remote operation
  _getopt.addOptionBool('r',"remote",false);
  _getopt.addOptionString('s',"server","http://edo.lcs.mit.edu/dclient/getwork.php");
  _getopt.addOptionString('u',"user","anonymous");
  _getopt.addOptionInt(0,"sleep",5);
  _getopt.addOptionInt(0,"retries",5);
  _getopt.addOptionInt(0,"proxy-port",80);
  _getopt.addOptionString(0,"proxy","");

  // for fetching a particular block from the server. requires authentication.
  _getopt.addOptionString(0,"wuid","");
  _getopt.addOptionString(0,"adminpw","");

  _getopt.addOptionBool('l',"local",false);
  _getopt.addOptionString('c',"ctext","96F8B204FD99534759A6C11A181EEDDFEB2DF1D4");
  _getopt.addOptionString('a',"alphabet","ABCDEFGHIJKLMNOPQRSTUVWXYZ 0123456789");
  _getopt.addOptionString('p',"pattern","??????");
  _getopt.addOptionString(0,"seed","");
  _getopt.addOptionString(0,"pattern-file","");

  _getopt.addOptionBool('b',"benchmark",false);
  _getopt.addOptionBool('t',"test",false);
}

MainApp::~MainApp()
{
}

int MainApp::run(int argc, char *argv[])
{
  _appname=argv[0];

  if (!_getopt.parse(argc, argv, true))
    {
      doHelp();
      return -1;
    }

  if (!doSelfTest())
    {
      return -1;
    }

  // if we were started in test mode, we're done now.
  if (_getopt.getOptionBool("test"))
    {
      return burnIn();
    }

  if (_getopt.getOptionBool("debug"))
    {
      _log.setLogLevel(Logger::DEBUG);
      printf("Options:\n\n");
      _getopt.showOptions();
      printf("\n");
    }

  if (_getopt.getOptionBool("silent"))
    {
      _getopt.setOptionBool("display-progress",false);
      _log.setConsole(NULL);
    }

  _log.setLogLevel(_getopt.getOptionInt("verbosity"));

  if (_getopt.getOptionBool("tivo32"))
    _getopt.setOptionString("ctext","96F8B204FD99534759A6C11A181EEDDFEB2DF1D4");

  if (_getopt.getOptionBool("tivo30"))
    _getopt.setOptionString("ctext","5CA5D9DBE5338BAB8690C79C9A9310BCD3A8F23B");
    
  if (strcmp(_getopt.getOptionString("log-file"),""))
    {
      char *path=_getopt.getOptionString("log-file");
      FILE *f=fopen(path,"a");
      if (f==NULL)
	{
	  perror(path);
	  return -1;
	}
      _log.setFile(f);
    }

  if (_getopt.getOptionBool("help"))
    {
      doHelp();
      return 0;
    }

  if (_getopt.getOptionBool("benchmark"))
    {
      doBenchmark();
      return 0;
    }

  if (_getopt.getOptionBool("local"))
    {
      doLocal();
      return 0;
    }

  if (_getopt.getOptionBool("remote"))
    {
      doRemote();
      return 0;
    }

  doHelp();
  return -1;
}

void MainApp::doHelp()
{
  printf("usage: %s [general-options] [command] [command-options]\n", _appname);
  printf("\n");
  printf("General Options:\n");
  printf("  -v | --verbosity LEVEL     Specify how verbose to be. (0-4, 4=most verbose)\n");
  printf("       --display-progress    Display progress information on stdout [default true]\n");
  printf("       --log-file PATHNAME   Direct messages to the specified log file.\n");
  printf("       --silent              Don't write anything to the console.\n");
  printf("\n");
  printf("Commands:\n");
  printf("  -r | --remote              Obtain work units from a remote server\n");
  printf("    -s | --server URL        URL of work server\n");
  printf("    -u | --user USER         userid to provide\n");
  printf("         --sleep MINUTES     How long to sleep for when an error occurs.\n");
  printf("         --retries #         Maximum number of retries before giving up.\n");
  printf("         --proxy             HTTP proxy server to use\n");
  printf("         --proxy-port        HTTP proxy port to use\n");
  printf("\n");
  printf("  -l | --local               Specify a work unit manually\n");
  printf("    -a | --alphabet ALPHA    Alphabet to use\n");
  printf("    -c | --ctext CTEXT       Ciphertext to use\n");
  printf("         --tivo32            Use tivo 3.2 ctext\n");
  printf("         --tivo30            Use tivo 3.0 ctext\n");
  printf("    -p | --pattern PATTERN   Pattern of search space [required]\n");
  printf("         --seed SEED         Seed to use with pattern.\n");
  printf("       | --pattern-file PATH Read patterns from the specified path.\n");
  printf("\n");
  printf("  -t | --test                Run in test mode\n");
  printf("  -h | --help                Shows this\n");
  printf("  -b | --benchmark           Determine speed of this machine\n");
  printf("\n");
  printf("Text in UPPERCASE indicates an argument to be provided\n");
  printf("by the user.\n\n");
  printf("Examples:\n\n");
  printf("%s -rs http://edo.lcs.mit.edu/dclient/getwork.php -u YOURUSERNAME\n",_appname);
  printf("  Connect to the server and process work units as 'eolson'.\n\n");
  printf("%s --displayprogress=false --verbose=false",_appname);
  printf("\n\n");
  
}

void MainApp::doBenchmark()
{
  struct timeval starttime, endtime;
  Cracker cracker;

  cracker.setAlphabet("ABCDEFGHIJKLMNOPQRSTUVWXYZ 0123456789");
  cracker.setCtext("5CA5D9DBE5338BAB8690C79C9A9310BCD3A8F23B");
  cracker.setProgressFrequency(_getopt.getOptionInt("progress-frequency"));

  _log.logf(Logger::HIGH,"Starting benchmark...\n");

  gettimeofday(&starttime, NULL);

  bool foundsolution=cracker.searchpattern("??????","3 ");

  gettimeofday(&endtime,NULL);

  if (!foundsolution)
    {
      printf("Benchmark failed; no solution was found!\n");
      printf("Do not run this client! It's broken!\n");
      return;      
    }

  char *sol=cracker.getSolution();
  if (sol==NULL || strcmp(sol,"3 0 BC"))
    {
      printf("Benchmark failed; the wrong solution was found!\n");
      printf("Do not run this client! It's broken!\n");
      return;
    }
  if (cracker.getProof()!=0xAE77AA35)
    {
      printf("Benchmark failed; we didn't generate the correct proof!\n");
      printf("Do not run this client! It's broken!\n");
      printf("got proof=%08X\n",cracker.getProof());
      return;
    }
  
  printf("[Internal tests passed]\n");

  // printf("Proof: %08X\n",cracker.getProof());

  long elapsedmsec=endtime.tv_sec*1000+endtime.tv_usec/1000
    -starttime.tv_sec*1000-starttime.tv_usec/1000;

  float t=elapsedmsec/1000.0;

  printf("Elapsed time: %.2f seconds\n",t);
  float tps=cracker.getCount()/t/1000.0;
  printf("%.1f kKeys/s\n",tps);

}

// these self tests take only a short while to run.
bool MainApp::doSelfTest()
{
  Cracker cracker;
  char *sol;
  int iter;

  cracker.setAlphabet("ABCDEFGHIJKLMNOPQRSTUVWXYZ 0123456789");
  cracker.setCtext("5CA5D9DBE5338BAB8690C79C9A9310BCD3A8F23B");
  cracker.setProgressFrequency(10);

  // try a known-to-fail pattern
  if (cracker.searchpattern("???",""))
    goto fail;

  if (cracker.getProof()!=0x83A5028D)
    {
      _log.logf(Logger::CRITICAL,"(got proof=%08X)\n",cracker.getProof());
      goto fail;
    }
  
  // try a known-to-work pattern
  if (!cracker.searchpattern("??????","3 0"))
    {
      goto fail;
    }

  sol=cracker.getSolution();

  if (sol==NULL || strcmp(sol,"3 0 BC"))
    {
      goto fail;
    }

  if (cracker.getProof()!=0x8C9EC3D4)
    {
      _log.logf(Logger::CRITICAL,"(got proof=%08X)\n",cracker.getProof());
      goto fail;
    }

  if (cracker.searchpattern("HAH?UQ ?N?FFN?? 00","3"))
    {
      goto fail;
    }
  
  if (cracker.getProof()!=0xEFA9756F)
    {
      _log.logf(Logger::CRITICAL,"(got proof=%08X)\n",cracker.getProof());
      goto fail;
    }
      
  _log.logf(Logger::LOW,"[Internal tests passed]\n");
  return true;

 fail:
  _log.logf(Logger::CRITICAL,"Internal tests failed; (wrong proof)\n");
  _log.logf(Logger::CRITICAL,"Do not run this client! It's broken!\n");
  return false;
}

bool MainApp::burnIn()
{
  Cracker cracker;
  char *sol;
  int iter;
  time_t starttime, endtime;
  int hours, minutes;

  starttime=time(NULL);

  _log.logf(Logger::MED,"Starting extended self test.");
  _log.logf(Logger::MED,"*** This test will run until cancelled (control-C).");
  _log.logf(Logger::MED,"    For a thorough test, you should let this run");
  _log.logf(Logger::MED,"    for at least 20 minutes.");

  iter=1;
  while (1)
    {
      _log.logf(Logger::MED,"Starting iteration %i",iter);

      doSelfTest();

      cracker.setAlphabet("ABCDEFGHIJKLMNOPQRSTUVWXYZ 0123456789");
      cracker.setCtext("5CA5D9DBE5338BAB8690C79C9A9310BCD3A8F23B");
      cracker.setProgressFrequency(4);
      
      if (cracker.searchpattern("?????",""))
	{
	  _log.logf(Logger::CRITICAL,"Found spurious solution\n");
	  goto fail;
	}
      
      if (cracker.getProof()!=0xF40E1244)
	{
	  _log.logf(Logger::CRITICAL,"(got proof=%08X)\n",cracker.getProof());
	  goto fail;
	}
      
      endtime=time(NULL);
      hours=(endtime-starttime)/3600;
      minutes=(endtime-starttime-hours*3600)/60;

      _log.logf(Logger::MED,"Passed. [%i hours, %i minutes elapsed total]", hours, minutes);
      iter++;
    }

 fail:
  _log.logf(Logger::CRITICAL,"Internal tests failed; (wrong proof)\n");
  _log.logf(Logger::CRITICAL,"Do not run this client! It's broken!\n");
  return false;
}

void MainApp::configureCracker(Cracker *cracker)
{
  cracker->enableProgress(_getopt.getOptionBool("display-progress"));
  cracker->setProgressFrequency(_getopt.getOptionInt("progress-frequency"));
}

bool MainApp::doPattern(Cracker *cracker, char *pattern, char *seed)
{
  _log.logf(Logger::MED,"Searching pattern '%s', seed '%s'.",pattern,seed);

  bool res=cracker->searchpattern(pattern,seed);
  
  if (res)
    {
      _log.logf(Logger::HIGH,"SOLUTION FOUND: '%s'",cracker->getSolution());
    }
  else
    _log.logf(Logger::MED,"Proof: %08X",cracker->getProof());

  return res;
}

void MainApp::doLocal()
{
  char *alphabet=_getopt.getOptionString("alphabet");
  char *pattern=_getopt.getOptionString("pattern");
  char *seed=_getopt.getOptionString("seed");
  char *ctext=_getopt.getOptionString("ctext");
  char *patternfile=_getopt.getOptionString("pattern-file");

  Cracker cracker;

  configureCracker(&cracker);

  cracker.setAlphabet(alphabet);
  cracker.setCtext(ctext);

  if (!strcmp(patternfile,""))
    {
      if (doPattern(&cracker, pattern,seed))
	return;

      Vector<char*> *extrapatterns=_getopt.getExtraArgs();
      for (int i=0;i<extrapatterns->getSize();i++)
	{
	  if (doPattern(&cracker,extrapatterns->get(i),seed))
	    return;
	}
    }
  else
    {
      FILE *f;

      if (!strcmp(patternfile,"-"))
	f=stdin;
      else
	f=fopen(patternfile,"r");

      if (f==NULL) 
	{
	  perror(patternfile);
	  return;
	}

      _log.logf(Logger::MED,"Testing patterns from path '%s', using seed '%s'.",patternfile,seed);
      char buf[32];
      while (fgets(buf,32,f))
	{
	  trim(buf);

	  if (doPattern(&cracker,buf,seed))
	    return;
	}	
    }
}

void MainApp::trim(char *b)
{
  int l=strlen(b);

  while (l>0 && isspace(b[l-1]))
    {
      b[l-1]=0;
      l--;
    }
}

void MainApp::doRemote()
{
  doRemoteThread();

}
void MainApp::doRemoteThread()
{
  char *server=_getopt.getOptionString("server");
  char *user=_getopt.getOptionString("user");

  while(1)
    {
      bool okay=doOneWorkUnit(server, user);

      // if they specified a work unit, only try once and then
      // stop.
      if (strcmp(_getopt.getOptionString("wuid"),""))
	return;

      if (!okay)
	{
	  // if some error occured, sleep for a while.
	  int sleepinterval=_getopt.getOptionInt("sleep");
	  if (sleepinterval<1)
	    sleepinterval=1;

	  _log.logf(Logger::HIGH,"Sleeping due to (hopefully transient) error.");
	  sleep(60*sleepinterval);
	}
    }
}


bool MainApp::doOneWorkUnit(char *getworkurl, char *user)
{
  ServerRequest sreq;
  char *requestwuid=_getopt.getOptionString("wuid");
  char *adminpw=_getopt.getOptionString("adminpw");

  Table *getworkargs=new Table();
  char *proxy=_getopt.getOptionString("proxy");
  int proxyport=_getopt.getOptionInt("proxy-port");

  if (strcmp(requestwuid,""))
    {
      getworkargs->put("WUID",requestwuid);
      getworkargs->put("ADMINPW",adminpw);
    }

  getworkargs->put("USERID",user);
  getworkargs->put("RUNTYPES","TC2");

  _log.logf(Logger::MED,"********************************************************");

  _log.logf(Logger::MED,"Requesting work unit.");

  if (strcmp(proxy,""))
    sreq.configureProxy(proxy,proxyport);

  Table *wu=sreq.performRequest(getworkurl, getworkargs);
  if (wu==NULL)
    {
      _log.logf(Logger::HIGH,"Error requesting work unit.");
      return false;
    }

  char *run=wu->get("RUN");
  char *alphabet=wu->get("ALPHABET");
  char *pattern=wu->get("PATTERN");
  char *seed=wu->get("SEED");
  char *ctext=wu->get("CTEXT");
  char *status=wu->get("STATUS");
  char *message=wu->get("MESSAGE");
  char *resulturl=wu->get("RESULTURL");
  char *nonce=wu->get("NONCE");
  char *wuid=wu->get("WUID");

  if (message!=NULL)
    _log.logf(Logger::CRITICAL, "SERVER SAID: %s",message);

  if (status!=NULL && !strcmp(status,"NOWORK"))
    {
      _log.logf(Logger::HIGH,"Server is out of work units.");
      return false;
    }

  if (wuid!=NULL)
    _log.logf(Logger::MED,"Processing work unit %s",wuid);

  if (status!=NULL && !strcmp(status,"FAIL"))
    {
      _log.logf(Logger::CRITICAL,"A critical problem occured while getting a work unit.");
      return false;
    }

  if (status!=NULL && !strcmp(status,"SHUTDOWN"))
    {
      _log.logf(Logger::CRITICAL, "Server told us to shutdown. Upgrade your client.",message);
      exit(0);
    }

  /* make sure we're not missing anything important */
  if (run==NULL || alphabet==NULL || pattern==NULL || ctext==NULL 
      || seed==NULL || status==NULL || resulturl==NULL || wuid==NULL
      || nonce==NULL)
    {
      _log.logf(Logger::HIGH,"Didn't get required data from server.");

      return false;
    }

  if (strcmp(run,"TC2"))
    {
      _log.logf(Logger::CRITICAL,"We can't perform this work unit.");
      return false;
    }

  Cracker cracker;
  configureCracker(&cracker);

  cracker.setAlphabet(alphabet);
  cracker.setCtext(ctext);
  
  bool res=doPattern(&cracker,pattern,seed);

  // submit the results.

  Table *results=new Table();
  results->put("USERID",user);
  results->put("CVER",CLIENTVERSION);
  results->put("WUID",wuid);
  results->put("NONCE",nonce);

  if (res)
    {
      results->put("SOLUTION",cracker.getSolution());
      results->put("PROOF","00000000");
    }
  else
    {
      char buf[32];
      snprintf(buf,32,"%08X",cracker.getProof());
      
      results->put("PROOF",buf);
    }

  results->put("STATUS","OKAY");

  bool okay=false;
  int tries=0;
  Table *resultresponse=NULL;
  int maxtries=_getopt.getOptionInt("retries");

  while (!okay && tries<maxtries)
    {
      resultresponse=sreq.performRequest(resulturl, results);
      if (resultresponse==NULL)
	{
	  _log.logf(Logger::HIGH,"Connection to server failed.");
	  okay=false;
	}
      else
	{
	  char *status=resultresponse->get("STATUS");
	  if (status!=NULL && !strcmp(status,"OKAY"))
	    {
	      okay=true;
	      break;
	    }
	}

      if (resultresponse!=NULL && resultresponse->get("MESSAGE")!=NULL)
	_log.logf(Logger::CRITICAL,"SERVER SAID: %s",resultresponse->get("MESSAGE"));

      okay=false;
      tries++;
      
      _log.logf(Logger::HIGH,"Can't submit work to server. Sleeping.");
      int minutes=_getopt.getOptionInt("sleep");
      if (minutes<1)
	minutes=1;
      sleep(minutes*60);

      if (resultresponse!=NULL)
	{
	  delete resultresponse;
	  resultresponse=NULL;
	}

    }
  
  if (!okay)
    _log.logf(Logger::CRITICAL,"Gave up trying to submit work to server.");
  else
    _log.logf(Logger::MED,"Work results submitted.");

  if (resultresponse!=NULL)
    delete resultresponse;
  delete getworkargs;
  delete wu;
  delete results;

  return true;

}


